#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Security.Permissions;
using System.ServiceModel;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Contracts.Interfaces;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.Server.Core.InternalInterfaces;
using HeuristicLab.Hive.Server.DataAccess;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Hive.Server.Core {
  [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerCall, ConcurrencyMode = ConcurrencyMode.Multiple)]
  public class SlaveFacade : ISlaveFacade {
    private ISlaveCommunicator slaveCommunicator = ServiceLocator.GetSlaveCommunicator();
    private IJobManager jobManager = ServiceLocator.GetJobManager();
    private IContextFactory contextFactory = ServiceLocator.GetContextFactory();

    public SlaveFacade() {
    }

    #region ISlaveCommunicator Members

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Response Login(SlaveDto slave) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.Login(slave);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseHeartBeat ProcessHeartBeat(HeartBeatData hbData) {
      using (contextFactory.GetContext(false)) { // due to concurrency issues with lifecycle-heartbeat do the transaction inside
        return slaveCommunicator.ProcessHeartBeat(hbData);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseObject<JobDto> GetJob(Guid slaveId) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.GetJob(slaveId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseResultReceived StoreFinishedJobResult(Guid slaveId, Guid jobId, byte[] result, TimeSpan executionTime, string exception) {
      using (contextFactory.GetContext()) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return slaveCommunicator.StoreFinishedJobResult(slaveId, jobId, result, executionTime, exception);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Response Logout(Guid slaveId) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.Logout(slaveId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Response IsJobStillNeeded(Guid jobId) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.IsJobStillNeeded(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseList<CachedHivePluginInfoDto> GetPlugins(List<HivePluginInfoDto> pluginList) {
      return slaveCommunicator.GetPlugins(pluginList);
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseResultReceived ProcessSnapshot(Guid slaveId, Guid jobId, byte[] result, TimeSpan executionTime, string exception) {
      using (contextFactory.GetContext()) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return slaveCommunicator.ProcessSnapshot(slaveId, jobId, result, executionTime, exception);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseCalendar GetCalendar(Guid slaveId) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.GetCalendar(slaveId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Response SetCalendarStatus(Guid slaveId, CalendarState state) {
      using (contextFactory.GetContext()) {
        return slaveCommunicator.SetCalendarStatus(slaveId, state);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseObject<JobDto> AddChildJob(Guid parentJobId, SerializedJob serializedJob) {
      using (contextFactory.GetContext()) {
        return jobManager.AddChildJob(parentJobId, serializedJob);
      }
    }
    #endregion

    #region ISlaveFacade Members

    /// <summary>
    /// Do not use automatic transactions here
    /// </summary>
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Stream GetStreamedJob(Guid slaveId) {
      ServiceLocator.GetLifecycleManager().JobsCurrentlyTransferring++;
      using (contextFactory.GetContext(false)) {
        ResponseObject<JobDto> job = null;
        MultiStream stream = new MultiStream();
        job = ServiceLocator.GetSlaveCommunicator().GetJob(slaveId);

        //first send response
        stream.AddStream(new StreamedObject<ResponseObject<JobDto>>(job));

        IInternalJobManager internalJobManager = (IInternalJobManager)ServiceLocator.GetJobManager();

        //second stream the job binary data
        MemoryStream memoryStream = new MemoryStream();
        if (job.Obj != null) {
          stream.AddStream(new MemoryStream(internalJobManager.GetSerializedJobDataById(job.Obj.Id)));
        }

        OperationContext slaveContext = OperationContext.Current;
        slaveContext.OperationCompleted += new EventHandler(delegate(object sender, EventArgs args) {
          if (stream != null) {
            stream.Dispose();
          }
          ServiceLocator.GetLifecycleManager().JobsCurrentlyTransferring--;
        });

        return stream;
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Stream GetStreamedPlugins(List<HivePluginInfoDto> pluginList) {
      return new StreamedObject<ResponseList<CachedHivePluginInfoDto>>(this.GetPlugins(pluginList));
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseResultReceived StoreFinishedJobResultStreamed(Stream stream) {
      using (contextFactory.GetContext()) {
        return ((IInternalSlaveCommunicator)slaveCommunicator).ProcessJobResult(stream, true);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseResultReceived ProcessSnapshotStreamed(Stream stream) {
      using (contextFactory.GetContext()) {
        return ((IInternalSlaveCommunicator)slaveCommunicator).ProcessJobResult(stream, false);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseObject<JobDto> PauseJob(SerializedJob serializedJob) {
      using (contextFactory.GetContext()) {
        return jobManager.PauseJob(serializedJob);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public JobResultList GetChildJobResults(Guid? parentJobId, bool recursive, bool includeParent) {
      using (contextFactory.GetContext(false)) {
        return jobManager.GetChildJobResults(parentJobId, recursive, includeParent).Obj;
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public SerializedJob GetLastSerializedResult(Guid jobId) {
      using (contextFactory.GetContext(false)) {
        return jobManager.GetLastSerializedResult(jobId).Obj;
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public Response DeleteChildJobs(Guid jobId) {
      using (contextFactory.GetContext()) {
        return jobManager.DeleteChildJobs(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Slave)]
    public ResponseObject<HivePluginFile> GetConfigurationFile() {
      return slaveCommunicator.GetConfigurationFile();
    }

    #endregion
  }
}
